/* Auto Plugin
 * Copyright (C) 1999 Adam Langley.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#define USE_PLUGIN

#include "../../src/common/xchat.h"
#include "../../src/common/plugin.h"
#include <regex.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>

#define	AUTO_TYPE_ONJOIN	1
#define	AUTO_TYPE_ONMSG		2
#define	AUTO_TYPE_ONPUBLIC	3

#define	AUTO_VERSION		"0.02"

#ifdef	ODD_BUF_CORRUPTION_THING
#define		malloc	mmap_malloc
#define		free	mmap_free

int	devzerofd, memsize;

void	*mmap_malloc (int size)
{
	void	*ret;
	
	ret = mmap(NULL, size + 4, PROT_READ | PROT_WRITE, MAP_PRIVATE, devzerofd, 0);
	if (ret == MAP_FAILED) {
		perror("mmap");
		exit(1);
	}
	memcpy (ret, &size, 4);
	
	memsize += size;
	
	return ret + 4;
}

void	mmap_free (void *buf)
{
	int	size;
		
	memcpy(&size, buf - 4, 4);
	munmap(buf - 4, size + 4);
	
	memsize -= size;
}

#endif

extern	struct module *module_find (char *name);

struct	auto_event	{
	char	type;
	int	eid;
	char	*nick;
	char	*chan;
	char	*msg;
	char	*action;
	struct	auto_event	*next, *last;
};

struct	auto_list_ent	{
	char	*data;
	struct	auto_list_ent	*next, *last;
};

struct	auto_list	{
	char	name[32];
	struct	auto_list_ent	*root;
	struct	auto_list	*next, *last;
};

int	auto_join_handler(struct server *serv, char *channel, char *user, char *ip, void *a, char b);
int	auto_lsevents (struct session *sess, char *tbuf, char *word[], char *word_eol[]);
int	auto_rmevent (struct session *sess, char *tbuf, char *word[], char *word_eol[]);
int	auto_autolist (struct session *sess, char *tbuf, char *word[], char *word_eol[]);
int	auto_on (struct session *sess, char *tbuf, char *word[], char *word_eol[]);
struct	auto_list *auto_get_list (char *name, char createflag);
int	auto_privmsg_handler(struct server *serv, char *user, char *ip, char *text, void *a, char b);
void	auto_purge_lists (struct auto_list *l);
void	auto_purge_list_ents (struct auto_list *l);
void	auto_purge_events (struct auto_event *e);
int	auto_writeconfig (struct session *sess, char *tbuf, char *word[], char *word_eol[]);
int	auto_readconfig (struct session *sess, char *tbuf, char *word[], char *word_eol[]);
int	auto_chanmsg_handler(struct server *serv, char *channel, char *from, char *text, void *, char fromme);

struct	commands	auto_cmds[] = {
	{"ON",		auto_on,	0, 0, "/ON <event> [...]\n"},
	{"LSEVENTS",	auto_lsevents,	0, 0, "/LSEVENTS <list>\n"},
	{"RMEVENT",	auto_rmevent,	0, 0, "/RMEVENT <list> <eid>\n"},
	{"AUTOLIST",	auto_autolist,	0, 0, "/AUTOLIST <list> <action> [<data>]\n"},
	{"AUTOSAVE",	auto_writeconfig,0,0, "/AUTOSAVE [<file>]\n"},
	{"AUTOREAD",	auto_readconfig,0, 0, "/AUTOREAD [<file>]\n"},
	{0, 0, 0, 0, 0}
};

char	*auto_event_errmsg = "Please give an events name, valid names are:\n\tjoin\n";
char	*auto_eventun_errmsg = "Unknown events name, valid names are:\n\tjoin\n";
char	*auto_mismsg_errmsg = "AutoDLL: Misfromed message\n";
char	*auto_nick2long_errmsg = "AutoDLL: Nick too long\n";
char	*auto_wrngtype_errmsg = "AutoDLL; Type wrong!\n";
char	*auto_openfile_errsmsg = "Error opening file\n";

char	*auto_default_fname = "xchat-auto.conf";

char	*auto_name = "Auto";
char	*auto_desc = "Auto is a simple auto op/kick/greet/ban module";
char	*AUTO_HEADER = "# Config file for the Xchat auto module\n# by Adam Langley\n\n";

char	*auto_act_nick, *auto_act_chan, *auto_act_msg, *auto_act_user;
struct	session		*auto_act_sess;

struct	module_cmd_set	auto_cmd_set;

/* JOIN */
struct	auto_event	*onjoinroot;
int	onjoineid = 0;

/* PRIVMSG */
struct	auto_event	*onmsgroot;
int	onmsgeid = 0;

/* PUBLIC */
struct	auto_event	*onpubroot;
int	onpubeid = 0;

struct	xp_signal	join_sig, privmsg_sig, public_sig;
int	(*join_next)	(void *, void *, void *, void *, void *, char);
int	(*privmsg_next)	(void *, void *, void *, void *, void *, char);
int	(*public_next)	(void *, void *, void *, void *, void *, char);

struct	auto_list	*auto_lists;

int	module_init (int ver, struct module *mod, struct session *sess)
{
	/* This check *MUST* be done first */
	if (ver != MODULE_IFACE_VER)
		return 1;
	
	if (module_find (auto_name) != NULL) {
		/* We are already loaded */
		PrintText(sess, "Module auto already loaded\n");
		return 1;
	}
	
	PrintText(sess, "\0035%%%\0037,5%%%\0038,7%%%\0035,8L O A D E D  A U T O  M O D U L E\0038,7%%%\0037,5%%%\0035,1%%%\n");
	PrintText(sess, "\0035%%%\0037,5%%%\0038,7%%%\0037,8L O A D E D  A U T O  M O D U L E\0038,7%%%\0037,5%%%\0035,1%%%\n");
	PrintText(sess, "\0035%%%\0037,5%%%\0038,7%%%\0038,8L O A D E D  A U T O  M O D U L E\0038,7%%%\0037,5%%%\0035,1%%%\n");
	PrintText(sess, "\0035%%%\0037,5%%%\0038,7%%%\0034,8L O A D E D  A U T O  M O D U L E\0038,7%%%\0037,5%%%\0035,1%%%\n");
	
	mod->name = auto_name;
	mod->desc = auto_desc;
	
	auto_cmd_set.mod = mod;
	auto_cmd_set.cmds = auto_cmds;
	
	module_add_cmds (&auto_cmd_set);
	
	join_sig.signal = XP_JOIN;
	join_sig.callback = XP_CALLBACK(auto_join_handler);
	join_sig.naddr = &join_next;
	join_sig.mod = mod;
	
	privmsg_sig.signal = XP_PRIVMSG;
	privmsg_sig.callback = XP_CALLBACK(auto_privmsg_handler);
	privmsg_sig.naddr = &privmsg_next;
	privmsg_sig.mod = mod;
	
	public_sig.signal = XP_CHANMSG;
	public_sig.callback = XP_CALLBACK(auto_chanmsg_handler);
	public_sig.naddr = &public_next;
	public_sig.mod = mod;
	
	hook_signal(&join_sig);
	hook_signal(&privmsg_sig);
	hook_signal(&public_sig);
	
	onjoinroot = NULL;
	onmsgroot = NULL;
	
	auto_lists = NULL;
	
#ifdef	ODD_BUF_CORRUPTION_THING
	memsize = 0;
	devzerofd = open("/dev/zero", O_RDONLY);
#endif
	
	return 0;
}

void	module_cleanup (struct module *mod, struct session *sess)
{
	PrintText(sess, "Unloading auto module\n");
	
	auto_purge_events (onjoinroot);
	auto_purge_events (onmsgroot);
	auto_purge_lists (auto_lists);
	
	module_rm_cmds(&auto_cmd_set);
}

void	auto_insert_event (struct auto_event **root, struct auto_event *new)
{
	new->next = *root;
	new->last = NULL;
	if (*root != NULL)
		(*root)->last = new;
	*root = new;
}

int	auto_onevent (struct session *sess, char *word[], char *word_eol[],
	char nickf, char msgf, char chanf, char actf, char type, int *eid,
	struct auto_event **root)
{
	int	nicklen, msglen, chanlen, actlen, totallen = 0;
	int	nickindex, msgindex, chanindex, actindex, n = 3;
	struct	auto_event	*new;
	
	if (nickf) nickindex = n++;
	if (chanf) chanindex = n++;
	if (msgf)  msgindex  = n++;
	if (actf)  actindex  = n++;
	
	if (nickf) nicklen = strlen(word[nickindex]);    else nicklen = -1;
	if (chanf) chanlen = strlen(word[chanindex]);    else chanlen = -1;
	if (msgf)  msglen  = strlen(word[msgindex]);     else msglen  = -1;
	if (actf)  actlen  = strlen(word_eol[actindex]); else actlen  = -1;
	
	if (nicklen == 0 || chanlen == 0 || msglen == 0 || actlen == 0)
		return TRUE;
	
	if (nickf) totallen += nicklen + 1;
	if (chanf) totallen += chanlen + 1;
	if (msgf)  totallen += msglen  + 1;
	if (actf)  totallen += actlen  + 1;
	
	new = malloc(sizeof(struct auto_event) + totallen);
	totallen = 0;
	if (nickf) { memcpy(new + sizeof(struct auto_event) + totallen, word[nickindex]    , nicklen + 1); totallen += nicklen + 1; }
	if (chanf) { memcpy(new + sizeof(struct auto_event) + totallen, word[chanindex]    , chanlen + 1); totallen += chanlen + 1; }
	if (msgf)  { memcpy(new + sizeof(struct auto_event) + totallen, word[msgindex]     , msglen  + 1); totallen += msglen  + 1; }
	if (actf)  { memcpy(new + sizeof(struct auto_event) + totallen, word_eol[actindex] , actlen  + 1); totallen += actlen  + 1; }
	totallen = 0;
	if (nickf) { new->nick   = (char *) (new + sizeof(struct auto_event) + totallen); totallen += nicklen + 1; } else new->nick   = NULL;
	if (chanf) { new->chan   = (char *) (new + sizeof(struct auto_event) + totallen); totallen += chanlen + 1; } else new->chan   = NULL;
	if (msgf)  { new->msg    = (char *) (new + sizeof(struct auto_event) + totallen); totallen += msglen  + 1; } else new->msg    = NULL;
	if (actf)  { new->action = (char *) (new + sizeof(struct auto_event) + totallen); totallen += actlen  + 1; } else new->action = NULL;

	new->type = type;
	new->eid  = *eid;
	(*(eid)) ++;
	
	auto_insert_event (root, new);
	
	return FALSE;
}

int	auto_on (struct session *sess, char *tbuf, char *word[], char *word_eol[])
{
	if (strlen(word[2]) == 0) {
		PrintText(sess, auto_event_errmsg);
		return TRUE;
	}
	
	if (strcasecmp(word[2], "join") == 0)
		return auto_onevent (sess, word, word_eol, 1, 0, 1, 1, AUTO_TYPE_ONJOIN, &onjoineid, &onjoinroot);
	if (strcasecmp(word[2], "msg") == 0)
		return auto_onevent (sess, word, word_eol, 1, 1, 0, 1, AUTO_TYPE_ONMSG, &onmsgeid, &onmsgroot);
	if (strcasecmp(word[2], "public") == 0)
		return auto_onevent (sess, word, word_eol, 1, 1, 1, 1, AUTO_TYPE_ONPUBLIC, &onpubeid, &onpubroot);
	
	PrintText(sess, auto_eventun_errmsg);
	return TRUE;
}	

void	auto_listevents (struct session *sess, struct auto_event *root)
{
	struct	auto_event	*cur;
	char	buf[1024];
	
	cur = root;
	for (;;) {
		if (cur == NULL)
			break;
		snprintf(buf, 1020, "%d\t%s, %s, %s, %s\n", cur->eid, cur->nick, cur->chan, cur->msg, cur->action);
		PrintText(sess, buf);
		cur = cur->next;
	}
}

int	auto_lsevents (struct session *sess, char *tbuf, char *word[], char *word_eol[])
{
	if (strlen(word[2]) == 0) {
		PrintText(sess, auto_event_errmsg);
		return TRUE;
	}
	
	if (strcasecmp(word[2], "join") == 0) {
		PrintText(sess, "JOIN Event List:\n");
		auto_listevents(sess, onjoinroot);
		return FALSE;
	}
	if (strcasecmp(word[2], "msg") == 0) {
		PrintText(sess, "MSG Event List:\n");
		auto_listevents(sess, onmsgroot);
		return FALSE;
	}
	if (strcasecmp(word[2], "public") == 0) {
		PrintText(sess, "PUBLIC Event List:\n");
		auto_listevents(sess, onpubroot);
		return FALSE;
	}
	if (strcasecmp(word[2], "all") == 0) {
		PrintText(sess, "JOIN Event List:\n");
		auto_listevents(sess, onjoinroot);
		PrintText(sess, "MSG Event List:\n");
		auto_listevents(sess, onmsgroot);
		PrintText(sess, "PUBLIC Event List:\n");
		auto_listevents(sess, onpubroot);
		return FALSE;
	}
	PrintText(sess, auto_eventun_errmsg);
	return TRUE;
}	

void	inline auto_unlinkevent (struct auto_event **root, struct auto_event *rm)
{
	if (rm->next != NULL)
		rm->next->last = rm->last;
	if (rm->last != NULL)
		rm->last->next = rm->next;
	if (rm == *root)
		*root = rm->next;
	free(rm);
}

void	inline auto_removeevent (struct session *sess, struct auto_event **root,
	int *eid, int id)
{
	struct	auto_event	*cur, *next;
	char	buf[512];
	
	cur = *root;
	for (;;) {
		if (cur == NULL)
			break;
		if (cur->eid == id) {
			snprintf(buf, 510, "Removing event %d\n", id);
			PrintText(sess, buf);
			next = cur->last;
			auto_unlinkevent (root, cur);
			break;
		}
		cur = cur->next;
	}
	if (cur == NULL) {
		snprintf(buf, 510, "Could not find event %d in event list\n", id);
		PrintText(sess, buf);
		return;
	}
	else {
		cur = next;
		*eid -= 1;
		for (;;) {
			if (cur == NULL)
				break;
			cur->eid--;
			cur = cur->last;
		}
	}
}

int	auto_rmevent (struct session *sess, char *tbuf, char *word[], char *word_eol[])
{
	if (strlen(word[2]) == 0) {
		PrintText(sess, auto_event_errmsg);
		return TRUE;
	}
	if (strlen(word[3]) == 0)
		return TRUE;
	
	if (strcmp(word[2], "join") == 0) {
		auto_removeevent(sess, &onjoinroot, &onjoineid, atoi(word[3]));
		return FALSE;
	}
	if (strcmp(word[2], "msg") == 0) {
		auto_removeevent(sess, &onmsgroot, &onmsgeid, atoi(word[3]));
		return FALSE;
	}
	if (strcmp(word[2], "public") == 0) {
		auto_removeevent(sess, &onpubroot, &onpubeid, atoi(word[3]));
		return FALSE;
	}
	if (strcmp(word[2], "all") == 0) {
		PrintText(sess, "Cannot do rmevent on the all list\n");
		return FALSE;
	}
}

void	auto_parse_list(char *list, void (*callback) (char *, char *, char *, char *),
	char	*nick, char *chan, char *action, char *msg)
{
	struct	auto_list	*l;
	struct	auto_list_ent	*e;
	
	l = auto_get_list(list, 0);
	if (l == NULL)
		return;
	
	e = l->root;
	for (;;) {
		if (e == NULL)
			break;
		if (nick == (char *) -1)
			callback(e->data, chan, action, msg);
		else if (chan == (char *) -1)
			callback(nick, e->data, action, msg);
		else if (action == (char *) -1)
			callback(nick, chan, action, msg);
		else if (msg == (char *) -1)
			callback(nick, chan, action, e->data);
		
		e = e->next;
	}
}

void	auto_do_action (char *action, char *nick, char *chan)
{
	char	buf[2048], *p, *end, pdibuf[2048], *word[32], *word_eol[32];
	int	n, nicklen, chanlen, len;
	
	if (chan != NULL)
		chanlen = strlen(chan);
	if (auto_act_msg != NULL)
		process_data_init(pdibuf, auto_act_msg, word, word_eol);

	end = strchr(nick, '!');
	if (end == NULL)
		nicklen = strlen(nick);
	else
		nicklen = end - nick;
	
	end = action + strlen(action) + 1;
	for (p = action, n = 0; p < end; p++) {
		if (*p != '%') {
			buf[n++] = *p;
			continue;
		}
		p++;
		switch (*p) {
		case '%':
			buf[n++] = '%';
			break;
		case 'n':
			memcpy(&buf[n], nick, nicklen);
			n += nicklen;
			break;
		case 'c':
			if (chan != NULL) {
				memcpy(&buf[n], chan, chanlen);
				n += chanlen;
			}
			break;
		case 'A':
			if (auto_act_msg != NULL) {
				p++;
				if (*p < 'A' || *p > 'Z')
					break;
				len = strlen(word[*p - 'A']);
				memcpy(&buf[n], word[*p - 'A'], len);
				n += len;
			}
			break;
		case 'a':
			if (auto_act_msg != NULL) {
				p++;
				if (*p < 'A' || *p > 'Z')
					break;
				len = strlen(word_eol[*p - 'A']);
				memcpy(&buf[n], word_eol[*p - 'A'], len);
				n += len;
			}
			break;
		}
	}
	
	buf[n] = 0;
	handle_command(buf, auto_act_sess, 0);
}

void	auto_event_callback(char *nick, char *chan, char *action, char *msg)
{
	struct	auto_list	*l;
	struct	auto_list_ent	*e;
	regex_t	pcomp;
	int	r;
	
	if (nick != NULL) {
		if (nick[0] == '$') {
			auto_parse_list (&nick[1], auto_event_callback, (char *) -1, chan, action, msg);
			return;
		}
	}
	if (msg != NULL) {
		if (msg[0] == '$') {
			auto_parse_list (&msg[1], auto_event_callback, nick, chan, action, (char *) -1);
			return;
		}
	}
	if (chan != NULL) {
		if (chan[0] == '$') {
			auto_parse_list (&chan[1], auto_event_callback, nick, (char *) -1, action, msg);
			return;
		}
	}
	
	if (msg != NULL) {
		if (regcomp (&pcomp, msg, REG_EXTENDED | REG_NOSUB) != 0)
			return;
		if (regexec(&pcomp, auto_act_msg, 0, NULL, 0) != 0)
			return;
		regfree(&pcomp);
	}

	if (nick != NULL) {
		if (regcomp (&pcomp, nick, REG_EXTENDED | REG_NOSUB) != 0)
			return;
		if (regexec(&pcomp, auto_act_nick, 0, NULL, 0) != 0)
			return;
		regfree(&pcomp);
	}
	
	if (chan != NULL) {
		if (regcomp (&pcomp, chan, REG_EXTENDED | REG_NOSUB) != 0)
			return;
		if (regexec(&pcomp, auto_act_chan, 0, NULL, 0) != 0)
			return;
		regfree(&pcomp);
	}
	
	if (action[0] != '$')
		auto_do_action(action, auto_act_user, auto_act_chan);
	
	l = auto_get_list (&action[1], 0);
	if (l == NULL)
		return;
	
	e = l->root;
	for (;;) {
		if (e == NULL)
			break;
		auto_do_action(e->data, auto_act_user, auto_act_chan);
		e = e->next;
	}
}		

int	auto_chanmsg_handler(struct server *serv, char *channel, char *from, char *text, void *a, char fromme)
{
	struct	auto_event	*e;	
	
	if (!fromme) {
		auto_act_user = auto_act_nick = from;
		auto_act_chan = channel;
		auto_act_msg = text;
		auto_act_sess = find_session_from_channel(channel, serv);
		if (auto_act_sess == NULL)
			auto_act_sess = serv->front_session;
	
		e = onpubroot;
	
		for (;;) {
			if (e == NULL)
				break;
			if (e->type != AUTO_TYPE_ONPUBLIC) {
				PrintText(auto_act_sess, auto_wrngtype_errmsg);
				break;
			}
			auto_event_callback(e->nick, e->chan, e->action, e->msg);
			
			e = e->next;
		}
	}
	
	XP_CALLNEXT(public_next, serv, channel, from, text, a, fromme);
}

int	auto_privmsg_handler(struct server *serv, char *user, char *ip, char *text, void *a, char b)
{
	char	nick[512];
	struct	auto_event	*e;
	
	memcpy(nick, user, strlen(user) + 1);
	strncat(nick, ip, 100);
	
	auto_act_nick = nick;
	auto_act_user = user;
	auto_act_chan = NULL;
	auto_act_sess = find_session_from_channel(user, serv);
	if (auto_act_sess == NULL)
		auto_act_sess = serv->front_session;
	auto_act_msg = text;
	
	e = onmsgroot;
	
	for (;;) {
		if (e == NULL)
			break;
		if (e->type != AUTO_TYPE_ONMSG) {
			PrintText(auto_act_sess, auto_wrngtype_errmsg);
			break;
		}
		auto_event_callback(e->nick, NULL, e->action, e->msg);
		
		e = e->next;
	}
	XP_CALLNEXT(privmsg_next, serv, user, ip, text, a, b);
}

int	auto_join_handler(struct server *serv, char *channel, char *user, char *ip, void *a, char b)
{
	char	nick[512];
	struct	auto_event	*e;
	
	memcpy(nick, user, strlen(user) + 1);
	strcat(nick, ip);
	
	e = onjoinroot;

	auto_act_nick = nick;
	auto_act_user = user;
	auto_act_chan = channel;
	auto_act_sess = find_session_from_channel(channel, serv);
	if (auto_act_sess == NULL)
		auto_act_sess = serv->front_session;
	
	for (;;) {
		if (e == NULL)
			break;
		if (e->type != AUTO_TYPE_ONJOIN) {
			PrintText(auto_act_sess, auto_wrngtype_errmsg);
			break;
		}
		auto_event_callback(e->nick, e->chan, e->action, NULL);

		e = e->next;
	}
	XP_CALLNEXT(join_next, serv, channel, user, ip, a, b);
}

/* *********** LIST HANDLING ************ */

void	inline auto_cannot_find_list_error (struct session *sess, 
	char *name, char *buf, int len)
{
	snprintf(buf, len, "Can not find list %s\n", name);
	PrintText(sess, buf);
}

void	inline auto_listrm_list (struct auto_list *l)
{
	if (l->next != NULL)
		l->next->last = l->last;
	if (l->last != NULL)
		l->last->next = l->next;
	if (l == auto_lists)
		auto_lists = l->next;
	free(l);
}

void	inline auto_listadd_list (struct auto_list *l)
{
	if (auto_lists != NULL)
		auto_lists->last = l;
	l->next = auto_lists;
	l->last = NULL;
	auto_lists = l;
}

struct	auto_list *auto_get_list (char *name, char createflag)
{
	struct	auto_list	*cur;
	
	cur = auto_lists;
	for (;;) {
		if (cur == NULL)
			break;
		if (strcmp(cur->name, name) == 0)
			return cur;
		cur = cur->next;
	}
	if (createflag) {
		cur = (struct auto_list *) malloc(sizeof(struct auto_list));
		memcpy(cur->name, name, strlen(name) + 1);
		cur->root = NULL;
		auto_listadd_list (cur);
		return cur;
	}
	else
		return NULL;
}

void	inline auto_entadd_list (struct auto_list *l, struct auto_list_ent *e, int pos)
{
	if (pos == -1) {
		if (l->root != NULL)
			l->root->last = e;
		e->next = l->root;
		e->last = NULL;
		l->root = e;
		return;
	}
}

void inline auto_entrm_list (struct auto_list *l, struct auto_list_ent *e)
{
	if (e->next != NULL)
		e->next->last = e->last;
	if (e->last != NULL)
		e->last->next = e->next;
	if (e == l->root)
		l->root = e->next;
	if (l->root == NULL)
		auto_listrm_list (l);
	free(e);
}

int	auto_list_ls (struct session *sess, char *word[], char *word_eol[])
{
	struct	auto_list	*l;
	struct	auto_list_ent	*e;
	char	buf[512];
	
	l = auto_get_list (word[3], 0);
	if (l == NULL) {
		auto_cannot_find_list_error(sess, word[3], buf, 510);
		return TRUE;
	}
	
	snprintf(buf, 510, "Contents of list %s:\n", word[3]);
	PrintText(sess, buf);
	e = l->root;
	for (;;) {
		if (e == NULL)
			break;
		snprintf(buf, 510, "\t%s\n", e->data);
		PrintText(sess, buf);
		e = e->next;
	}
	
	return FALSE;
}

int	auto_list_rm (struct session *sess, char *word[], char *word_eol[])
{
	struct	auto_list	*l;
	struct	auto_list_ent	*e;
	int	entlen;
	char	buf[512];
	
	entlen = strlen(word_eol[4]);
	if (entlen == 0) {
		PrintText(sess, "Usage: /AUTOLIST rm <list> <data>\n");
		return FALSE;
	}
	l = auto_get_list (word[3], 0);
	if (l == NULL) {
		auto_cannot_find_list_error(sess, word[3], buf, 510);
		return FALSE;
	}
	
	e = l->root;
	for (;;) {
		if (e == NULL)
			break;
		if (strcmp(e->data, word_eol[4]) == 0) {
			auto_entrm_list (l, e);
			return FALSE;
		}
		e = e->next;
	}
	
	PrintText(sess, "No matching entry in list\n");
	return FALSE;
}	

int	auto_list_add (struct session *sess, char *word[], char *word_eol[], int pos)
{
	struct	auto_list	*l;
	struct	auto_list_ent	*e;
	int	entlen;
	
	entlen = strlen(word_eol[4]);
	if (entlen == 0) {
		PrintText(sess, "Usage: /AUTOLIST add <list> <data>\n");
		return FALSE;
	}
	l = auto_get_list (word[3], 1);

	e = (struct auto_list_ent *) malloc(sizeof(struct auto_list) + entlen + 1);

	memcpy(e + sizeof(struct auto_list), word_eol[4], entlen + 1);
	e->data = (char *) (e + sizeof(struct auto_list));

	auto_entadd_list (l, e, pos);
		
	return FALSE;
}

void	auto_clear_list (struct auto_list *l)
{
	struct	auto_list_ent	*e, *nexte;
	
	e = l->root;
	for (;;) {
		if (e == NULL)
			break;
		nexte = e->next;
		free(e);
		e = nexte;
	}
}

int	auto_list_clear (struct session *sess, char *word[], char *word_eol[])
{
	struct	auto_list	*l;
	char	buf[512];
	
	l = auto_get_list (word[3], 0);
	if (l == NULL) {
		auto_cannot_find_list_error(sess, word[3], buf, 510);
		return FALSE;
	}
	auto_clear_list (l);
	auto_listrm_list (l);
}

int	auto_autolist (struct session *sess, char *tbuf, char *word[], char *word_eol[])
{
	if (strlen(word[2]) == 0 || strlen(word[3]) == 0)
		return TRUE;
	if (strlen(word[3]) > 31) {
		PrintText (sess, "List name too long (max 31 chars)\n");
		return TRUE;
	}
		
	if ( strcasecmp(word[2], "add") == 0)
		return auto_list_add(sess, word, word_eol, -1);
	if ( strcasecmp(word[2], "rm") == 0)
		return auto_list_rm (sess, word, word_eol);
	if ( strcasecmp(word[2], "ls") == 0)
		return auto_list_ls (sess, word, word_eol);
	if ( strcasecmp(word[2], "clear") == 0)
		return auto_list_clear (sess, word, word_eol);
	
	PrintText(sess, "Unknown command, valid commands are:\n\tadd\n\trm\n\tls\n\tclear\n\n");
	return FALSE;
}

/* ******************** PURGE CODE ************************ */

void	auto_purge_events (struct auto_event *e)
{
	struct	auto_event	*next, *cur;
	
	cur = e;
	for (;;) {
		if (cur == NULL)
			return;
		next = cur->next;
		free(cur);
		cur = next;
	}
}

void	auto_purge_list_ents (struct auto_list *l)
{
	struct	auto_list_ent	*next, *cur;
	
	cur = l->root;
	for (;;) {
		if (cur == NULL)
			return;
		next = cur->next;
		free(cur);
		cur = next;
	}
}

void	auto_purge_lists (struct auto_list *l)
{
	struct	auto_list	*next, *cur;
	
	cur = l;
	for (;;) {
		if (cur == NULL)
			return;
		next = cur->next;
		auto_purge_list_ents(cur);
		free(cur);
		cur = next;
	}
}

/* ******************* FILE HANDLING *******************/
/* *********************** WRITE **********************/


void	auto_write_list_ents (int fd, struct auto_list_ent *cur)
{
	char	tab[] = "    ";
	char	newline[] = "\n";

	for (;;) {
		if (cur == NULL)
			return;
		write(fd, tab, sizeof(tab) - 1);
		write(fd, cur->data, strlen(cur->data));
		write(fd, newline, sizeof(newline) - 1);
		
		cur = cur->next;
	}
}

void	auto_write_events_ent (int fd, struct auto_event *e)
{
	char	sep[] = "# ********************************\n\n";
	char	null[] = "NULLENTRY\n";
	char	newline[] = "\n";
	char	tab[]=  "    ";

	for (;;) {
		if (e == NULL)
			return;
			
		if (e->nick == NULL) {
			write(fd, tab, sizeof(tab) - 1);
			write(fd, null, sizeof(null) - 1);
		}
		else {
			write(fd, tab, sizeof(tab) - 1);
			write(fd, e->nick, strlen(e->nick));
			write(fd, newline, sizeof(newline) - 1);
		}

		if (e->chan == NULL) {
			write(fd, tab, sizeof(tab) - 1);
			write(fd, null, sizeof(null) - 1);
		}
		else {
			write(fd, tab, sizeof(tab) - 1);
			write(fd, e->chan, strlen(e->chan));
			write(fd, newline, sizeof(newline) - 1);
		}

		if (e->msg == NULL) {
			write(fd, tab, sizeof(tab) - 1);
			write(fd, null, sizeof(null) - 1);
		}
		else {
			write(fd, tab, sizeof(tab) - 1);
			write(fd, e->msg, strlen(e->msg));
			write(fd, newline, sizeof(newline) - 1);
		}

		if (e->action == NULL) {
			write(fd, tab, sizeof(tab) - 1);
			write(fd, null, sizeof(null) - 1);
		}
		else {
			write(fd, tab, sizeof(tab) - 1);
			write(fd, e->action, strlen(e->action));
			write(fd, newline, sizeof(newline) - 1);
		}
		
		write(fd, sep, sizeof(sep) - 1);
		e = e->next;
	}
}

void	auto_write_events (int fd)
{
	char	openevents[] = "# ************** EVENTS ************\n# These are in the form of a series of 4 lines in the order nick,chan,msg,action, with $$NULL where needed\n\n";
	char	closeevents[] = "# ********** END OF EVENTS ********\n\n";
	char	newevent[] = "$$EVENTS\n";
	char	endevent[] = "$$ENDEVENTS\n\n";
	char	newline[] = "\n";
	
	char	join[] = "join\n";
	char	msg[] = "msg\n";
	char	joinbanner[] = "# *********** JOIN EVENTS ***********\n";
	char	msgbanner[] = "# ************* MSG EVENTS ***********\n";
	
	write (fd, openevents, sizeof(openevents) - 1);
	
	if (onjoinroot != NULL) {
		write (fd, joinbanner, sizeof(joinbanner) - 1);
		write (fd, newevent, sizeof(newevent) - 1);
		write (fd, join, sizeof(join) - 1);
		write(fd, newline, sizeof(newline) - 1);
		auto_write_events_ent (fd, onjoinroot);
		write (fd, endevent, sizeof(endevent) - 1);
	}
	
	if (onmsgroot != NULL) {
		write (fd, msgbanner, sizeof(msgbanner) - 1);
		write (fd, newevent, sizeof(newevent) - 1);
		write (fd, msg, sizeof(msg) - 1);
		write(fd, newline, sizeof(newline) - 1);
		auto_write_events_ent (fd, onmsgroot);
		write (fd, endevent, sizeof(endevent) - 1);
	}

	write (fd, closeevents, sizeof(closeevents) - 1);
}
	
void	auto_write_lists (int fd, struct auto_list *cur)
{
	char	openlist[] = "# ********* LISTS *********\n\n";
	char	closelist[] = "# ****** END OF LISTS ****\n\n"; 
	char	newlist[] = "$$LIST\n";
	char	endlist[] = "$$ENDLIST\n";
	char	newline[] = "\n";

	write(fd, openlist, sizeof(openlist) - 1);
	for (;;) {
		if (cur == NULL)
			break;
		write(fd, newlist, sizeof(newlist) - 1);
		write(fd, cur->name, strlen(cur->name));
		write(fd, newline, sizeof(newline) - 1);
		write(fd, newline, sizeof(newline) - 1);
		auto_write_list_ents(fd, cur->root);
		write(fd, endlist, sizeof(endlist) - 1);
		
		cur = cur->next;
	}
	
	write(fd, closelist, sizeof(closelist) - 1);
}

int	auto_writeconfig (struct session *sess, char *tbuf, char *word[], char *word_eol[])
{
	char	*fname, *header = AUTO_HEADER;
	char	buf[512];
	int	fd, headerlen = strlen(AUTO_HEADER);
	
	if (strlen(word[2]) != 0)
		fname = word[2];
	else
		fname = auto_default_fname;
	
	snprintf(buf, 510, "Saving configuration information to %s\n", fname);
	PrintText(sess, buf);
	
	fd = open(fname, O_WRONLY | O_TRUNC | O_CREAT, 0x1a8);
	if (fd < 0) {
		PrintText(sess, auto_openfile_errsmsg);
		return TRUE;
	}
	
	write (fd, header, headerlen);
	
	auto_write_lists(fd, auto_lists);
	auto_write_events(fd);
	
	close(fd);
}

/* ******************* FILE HANDLING *******************/
/* *********************** READ ************************/

int	inline auto_read_line (char *buf, int blen, char *map, int *n, int len)
{
	char	*p = &map[*n];
	int	num = *n, i;
	
	for (i = 0; num <= len; num++, p++, i++) {
		switch (*p) {
		case '\n':
			buf[i] = 0;
			*n = ++num;
			return 0;
		default:
			buf[i] = *p;
			if (i == blen) {
				buf[blen] = 0;
				*n = num;
				return 1;
			}
			break;
		}
	}
	buf[i] = 0;
	return 2;
}

int	auto_read_strip (char *buf)
{
	int	i, len = strlen(buf);
	
	for (i = 0; i < len; i++) {
		if (buf[i] != ' ' && buf[i] != '\t')
			break;
	}
	return i;
}

#define	STATUS_NEWLIST	1
#define	STATUS_INLIST	2
#define	STATUS_OPEN	3
#define	STATUS_NEWEVENT	4
#define STATUS_INEVENT	5
#define	SSTATUS_INEVENTN	1
#define	SSTATUS_INEVENTC	2
#define	SSTATUS_INEVENTM	3
#define	SSTATUS_INEVENTA	4

int	auto_read_loop (struct session *sess, char *map, int len)
{
	char	buf[2048], *line, cmd[128], e_type[2048], e_nick[2048], e_chan[2048], 
		e_msg[2048], e_action[2048];
	int	p = 0, s, r, eof = 0, llen, *eid;
	int	status = STATUS_OPEN, sub_status;
	
	struct	auto_list	*list;
	struct	auto_list_ent	*listent, *lastlistent;
	
	struct	auto_event	*events, *lastevents, **root;
	
	while (!eof) {
		r = auto_read_line (buf, 2048, map, &p, len);
		if (r != 0) {
			if (r == 1) {
				PrintText(sess, "Line too long while reading config file; skipping: configuration corrupt\n");
				continue;
			}
			else
				eof = 1;
		}
		s = auto_read_strip(buf);
		line = &buf[s];
		llen = strlen(line);
		if (llen == 0)
			continue;
		if (line[0] == '#')
			continue;
		
		if (memcmp(line, "$$", 2) == 0) {
			/* Esacpe command */
			if (llen > (sizeof(cmd) - 2))
				return 1;
			memcpy(cmd, &line[2], llen - 2);
			cmd[llen - 2] = 0;
			
			if (strcasecmp(cmd, "LIST") == 0) {
				if (status != STATUS_OPEN)
					PrintText(sess, "$$LIST unexpected\n");
				status = STATUS_NEWLIST;
				continue;
			}
			if (strcasecmp(cmd, "ENDLIST") == 0) {
				if (status != STATUS_INLIST)
					PrintText(sess, "$$ENDLIST while not in list\n");
				status = STATUS_OPEN;
				continue;
			}
			if (strcasecmp(cmd, "EVENTS") == 0) {
				if (status != STATUS_OPEN)
					PrintText(sess, "$$EVENTS unexpected\n");
				status = STATUS_NEWEVENT;
				continue;
			}
			if (strcasecmp(cmd, "ENDEVENTS") == 0) {
				if (status != STATUS_INEVENT)
					PrintText(sess, "$$ENDEVENTS while not in event\n");
				status = STATUS_OPEN;
				continue;
			}
			continue;
		}
		
		switch (status) {
		case STATUS_NEWLIST:
			list = auto_get_list(line, 1);
			lastlistent = NULL;
			status = STATUS_INLIST;
			continue;

		case STATUS_INLIST:
			listent = (struct auto_list_ent *) malloc (
					sizeof(struct auto_list_ent) + len + 1);
			memcpy(listent + sizeof(struct auto_list_ent), line, len + 1);
			listent->data = (char *) (listent + sizeof(struct auto_list_ent));
			listent->next = NULL;
			listent->last = lastlistent;
			if (lastlistent != NULL)
				lastlistent->next = listent;
			else
				list->root = listent;
			lastlistent = listent;
			continue;
		case STATUS_NEWEVENT:
			lastevents = NULL;
			status = STATUS_INEVENT;
			sub_status = SSTATUS_INEVENTN;
			strcpy(e_type, line);
			continue;
		case STATUS_INEVENT:
			switch(sub_status) {
			case SSTATUS_INEVENTN:
				sub_status = SSTATUS_INEVENTC;
				if (strcmp(line, "NULLENTRY") == 0)
					strcpy(e_nick, "");
				else
					strcpy(e_nick, line);
				continue;
			case SSTATUS_INEVENTC:
				sub_status = SSTATUS_INEVENTM;
				if (strcmp(line, "NULLENTRY") == 0)
					strcpy(e_chan, "");
				else
					strcpy(e_chan, line);
				continue;
			case SSTATUS_INEVENTM:
				sub_status = SSTATUS_INEVENTA;
				if (strcmp(line, "NULLENTRY") == 0)
					strcpy(e_msg, "");
				else
					strcpy(e_msg, line);
				continue;
			case SSTATUS_INEVENTA:
				if (strcmp(line, "NULLENTRY") == 0)
					strcpy(e_action, "");
				else
					strcpy(e_action, line);
				llen = strlen(e_action);
			
				sub_status = SSTATUS_INEVENTN;
				events = (struct auto_event *) malloc (sizeof(struct auto_event) +
					strlen(e_nick) + strlen(e_chan) + strlen(e_msg) + llen + 4);
				memcpy(events + sizeof(struct auto_event), e_nick, strlen(e_nick) + 1);
				events->nick = (char *) (events + sizeof(struct auto_event));
				memcpy(events + sizeof(struct auto_event) + strlen(e_nick) + 1, e_chan, 
					strlen(e_chan) + 1);
				events->chan = (char *)(events + sizeof(struct auto_event) + strlen(e_nick) + 1);
				memcpy(events + sizeof(struct auto_event) + strlen(e_nick) + strlen(e_chan) + 2, 
					e_msg, strlen(e_msg) + 1);
				events->msg = (char *) (events + sizeof(struct auto_event) + strlen(e_nick) + strlen(e_chan) + 2);
				memcpy(events + sizeof(struct auto_event) + strlen(e_nick) + strlen(e_chan) + strlen(e_msg) + 3,
					e_action, llen + 1);
				events->action = (char *) (events + sizeof(struct auto_event) +
					strlen(e_nick) + strlen(e_chan) + strlen(e_msg) + 3);
				
				if (strcmp(e_type, "join") == 0) {
					events->type = AUTO_TYPE_ONJOIN;
					eid = &onjoineid;
					root = &onjoinroot;
				}
				else if (strcmp(e_type, "msg") == 0) {
					events->type = AUTO_TYPE_ONMSG;
					eid = &onmsgeid;
					root = &onmsgroot;
				}
				else {
					printf("Unknown event type %s\n", e_type);
					exit(1);
				}
				events->eid = *eid;
				(*eid)++;
				if (lastevents != NULL)
					lastevents->next = events;
				else
					*root = events;
				lastevents = events;
				events->next = NULL;
				continue;
			}
		}
	}
}

int	auto_readconfig (struct session *sess, char *tbuf, char *word[], char *word_eol[])
{
	char	*fname, *map;
	char	buf[512];
	int	fd;
	struct	stat	st;
	
	if (strlen(word[2]) != 0)
		fname = word[2];
	else
		fname = auto_default_fname;
	
	snprintf(buf, 510, "Loading configuration information to %s\n", fname);
	PrintText(sess, buf);
	
	fd = open(fname, O_RDONLY);
	if (fd < 0) {
		PrintText(sess, auto_openfile_errsmsg);
		return TRUE;
	}
	
	if (fstat(fd, &st) != 0) {
		PrintText(sess, auto_openfile_errsmsg);
		return TRUE;
	}

	map = mmap(NULL, st.st_size, PROT_READ, MAP_PRIVATE, fd, 0);
	if (map == MAP_FAILED) {
		PrintText(sess, auto_openfile_errsmsg);
		return TRUE;
	}

	auto_purge_events (onjoinroot);
	auto_purge_events (onmsgroot);
	auto_purge_lists (auto_lists);
	onjoinroot = NULL;
	onmsgroot = NULL;
	onjoineid = 0;
	onmsgeid = 0;
	auto_lists = NULL;

	auto_read_loop(sess, map, st.st_size);
	munmap(map, st.st_size);
	close(fd);
}
